from abc import ABC, abstractmethod
from typing import Any, Callable, Coroutine, Type, TypeVar, Union

from auto_translategram.auto_translategram.cache import Cache
from auto_translategram.auto_translategram.translator_services import TranslatorService

T = TypeVar('T')


class Translator(ABC):
    """
    Abstract base class for implementing translation functionality in various frameworks.
    This class provides a uniform interface for translation, allowing adapters to be written for different frameworks.
    This class is meant to be subclassed, and the `handler_translator` method should be implemented in the subclass.
    """
    def __init__(self, translator_service: TranslatorService, cache_system: Union[Type[Cache], None] = None) -> None:
        """
        Initializes a new Translator instance using the specified `translator_service`.

        :param translator_service: The `BaseTranslatorService` to use for translations.
        """
        ...

    @abstractmethod
    def handler_translator(
        self,
        message: str
    ) -> Callable[[Callable[..., T]], Callable[[Any, Any, str], Coroutine[Any, Any, Any]]]:
        """
        Translate a message based on the users' language
        :param func: The handler function that is used for handling commands by Frameworks.
        :param message: The message to translate.
        """
        ...



from typing import Protocol
from auto_translategram.auto_translategram.service_libs import mtranslate


class TranslatorService(Protocol):
    """
    Defines the interface for the translator services
    """
    async def translate_str(self, text: str, target_language: str = 'auto', source_language: str = 'auto') -> str:
        """
        Returns the translated string
        """
        ...


class MtranslateTranslatorService:
    """
    Implements the BaseTranslatorService protocol using the mtranslate library
    """
    def __init__(self) -> None:
        assert mtranslate, '`MtranslateTranslatorService` requires `mtranslate` package'
        self.service = mtranslate

    async def translate_str(self, text: str, target_language: str = 'auto', source_language: str = 'auto') -> str:
        """
        Returns the translated string using the mtranslate library
        """
        if not isinstance(text, str) or not isinstance(target_language, str) or not isinstance(source_language, str):
            raise TypeError('`text`, `target_language` and `source_language` must be a string')
        translated_text = self.service.translate(
            to_translate=text,
            to_language=target_language,
            from_language=source_language
            )
        return str(translated_text)



from auto_translategram.python_telegram_bot_translator.adapter import PythonTelegramBotAdapter as PythonTelegramBotTranslator
from auto_translategram.auto_translategram.translator_services import MtranslateTranslatorService
from auto_translategram.auto_translategram.cache import PickleCache



import inspect
from typing import Any, Coroutine, Callable, Type, TypeVar, Union
from telegram.ext import ContextTypes

from telegram import Update

from auto_translategram.auto_translategram.cache import Cache
from auto_translategram.auto_translategram.translator_services import TranslatorService
from auto_translategram.auto_translategram.translator import Translator

_T = TypeVar('_T')


class PythonTelegramBotAdapter(Translator):
    """
    A Translator adapter for the python-telegram-bot framework.

    Inherits from the abstract class `Translator` and implements its `handler_translator` method
    to provide translation functionality for the python-telegram-bot framework.

    :param translator_service: The `TranslatorService` to use for translations.
    """

    def __init__(self,
                 translator_service: Type[TranslatorService],
                 cache_system: Union[Type[Cache], None] = None
                 ) -> None:
        """
        Initializes a new PythonTelegramBotAdapter instance using the specified `translator_service`.

        :param translator_service: The `TranslatorService` to use for translations.
        """
        self._translator_service = translator_service()
        self._cache_system = cache_system

    def handler_translator(
            self,
            message: str
            ) -> Callable[[Callable[[Any, Any, str], _T]], Callable[[Any, Any, str], Coroutine[Any, Any, Any]]]:
        """
        A decorator that wraps a python-telegram-bot `handler` function to provide translation functionality.

        The decorated function will be executed after being wrapped with a new function that translates
        the incoming message into the user's preferred language (if it is not already in that language).
        If the user does not have a preferred language set or if it is set to 'en', the message will not be translated.

        :param func: The handler function that is used for handling commands by the Python-telegram-bot framework.
        :param message: The message to translate.
        :return: A coroutine that wraps the handler function and provides translation functionality.
        """
        def decorator(func: Callable[[Update, ContextTypes.DEFAULT_TYPE, str], _T]) \
                -> Callable[[Any, Any, str], Coroutine[Any, Any, Any]]:

            async def wrapper(update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = message) -> Any:
                user_lang = update.effective_user.language_code if update.effective_user else 'en'
                message = message
                msg = message
                if user_lang != 'en' and user_lang is not None:  # TODO: get rid of this!
                    if self._cache_system is not None:
                        msg = await self._cache_system.retrieve(
                            key=func.__name__ + '_' + user_lang
                            )  # type: ignore
                        if msg is None:
                            msg = await self._translator_service.translate_str(
                                text=message,
                                target_language=user_lang,
                                source_language='en'
                                )
                            await self._cache_system.store(
                                key=func.__name__ + '_' + user_lang, value=msg
                                )  # type: ignore
                    else:
                        msg = await self._translator_service.translate_str(
                            text=message,
                            target_language=user_lang,
                            source_language='en'
                            )
                is_async = inspect.iscoroutinefunction(func)
                if is_async:
                    result = await func(update, context, str(msg))  # type: ignore[misc]
                else:
                    result = func(update, context, str(msg))  # type: ignore[misc]
                return result

            return wrapper
        return decorator



import pickle
from typing import Protocol, TypeVar, Union

T = TypeVar('T')


class Cache(Protocol):

    async def store(self, key: str, value: str) -> None:
        ...

    async def retrieve(self, key: str) -> Union[str, None]:
        ...


class PickleCache:
    def __init__(self, obj: T, filename: str = "translation.data") -> None:
        self._obj = obj
        self.pickle_file = filename
        with open(self.pickle_file, 'ab') as file:
            pickle.dump(self._obj, file)

    async def store(self, key: str, value: str) -> None:
        setattr(self._obj, key, value)
        with open(self.pickle_file, 'wb') as file:
            pickle.dump(self._obj, file)

    async def retrieve(self, key: str) -> Union[str, None]:
        with open(self.pickle_file, 'rb') as file:
            loaded_data = pickle.load(file)
        return loaded_data.__dict__.get(key) if isinstance(loaded_data.__dict__.get(key), str) else None



from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='translategram',
    version='0.2.0',
    description='Translategram is a Python library for translating messages are sent by your Bot in Telegram',
    url='https://github.com/EkberHasanov/translategram',
    author='Akbar',
    author_email='hasanvakbar@gmail.com',
    license='MIT',
    install_requires=[
        'mtranslate',
    ],
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["tests", "tests.*", "*.tests", "*.tests.*"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)
